/*
 * BSD LICENSE
 *
 * Copyright(c) 2025-2026 Intel Corporation. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of Intel Corporation nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * Dump module
 */

#ifndef __DUMP_RMID_H__
#define __DUMP_RMID_H__

#ifdef __cplusplus
extern "C" {
#endif

#include "pqos.h"

/**
 * @brief Selects domain ids for --dump-rmid-regs command
 *
 * @param arg Domain IDs string from --dump-rmid-domain-ids option
 */
void selfn_dump_rmid_domain_ids(const char *arg);

/**
 * @brief Selects rmids for --dump-rmid-regs command
 *
 * @param arg RMIDs string from --dump-rmids option
 */
void selfn_dump_rmids(const char *arg);

/**
 * @brief Selects memory regions in --dump-rmid-regs command
 *
 * @param arg Memory Regions string from --dump-rmid-mem-regions option
 */
void selfn_dump_rmid_mem_regions(const char *arg);

/**
 * @brief Selects rmid type for --dump-rmid-regs command
 *
 * @param arg RMID type string from --dump-rmid-type option
 */
void selfn_dump_rmid_type(const char *arg);

/**
 * @brief Selects binary option in --dump-rmid-regs command
 *
 * @param arg not used
 */
void selfn_dump_rmid_binary(const char *arg);

/**
 * @brief Selects upscale option in --dump-rmid-regs command
 *
 * @param arg not used
 */
void selfn_dump_rmid_upscaling(const char *arg);

/**
 * @brief Prints the values of RMID registers for the --dump-rmid-regs command
 *
 * @param [in] sys PQoS system configuration struct
 *                 returned by a pqos_sysconfig_get
 */
void dump_rmid_regs(const struct pqos_sysconfig *sys);

#ifdef __cplusplus
}
#endif

#endif /* __DUMP_RMID_H__ */
